from __future__ import annotations

import enum
from contextlib import contextmanager
from typing import TYPE_CHECKING, BinaryIO

if TYPE_CHECKING:
    from collections.abc import Iterator


class ThumbnailType(enum.IntEnum):
    WINDOWS_VISTA = 0x14
    WINDOWS_7 = 0x15
    WINDOWS_81 = 0x1F
    WINDOWS_10 = 0x20


@contextmanager
def seek_and_return(fh: BinaryIO, position: int) -> Iterator[BinaryIO]:
    current_position = fh.tell()
    try:
        fh.seek(position)
        yield fh
    finally:
        fh.seek(current_position)



#!/bin/python3
from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.thumbcache.exceptions import Error
from dissect.thumbcache.thumbcache_file import ThumbcacheFile
from dissect.thumbcache.tools.utils import create_argument_parser, write_entry

if TYPE_CHECKING:
    from pathlib import Path


def dump_entry_data(path: Path, output_dir: Path) -> None:
    with path.open("rb") as file:
        try:
            cache_file = ThumbcacheFile(file)

            for entry in cache_file.entries():
                write_entry(output_dir, entry, path.name.split("_", 1)[1])
        except Error as e:
            print(e)


def main() -> None:
    parser = create_argument_parser("extract raw thumbcache entries")
    args = parser.parse_args()
    path: Path = args.cache_path

    if path.name.endswith("idx.db"):
        parser.exit("Please provide a file other than the idx file.")

    if path.is_dir():
        for file in path.iterdir():
            if file.name.endswith("idx.db"):
                continue
            dump_entry_data(file, args.output_dir)
    else:
        dump_entry_data(path, args.output_dir)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.thumbcache.index import ThumbnailIndex
from dissect.thumbcache.thumbcache_file import ThumbcacheFile

if TYPE_CHECKING:
    from collections.abc import Iterator
    from pathlib import Path

    from dissect.thumbcache.index import IndexEntry
    from dissect.thumbcache.thumbcache_file import ThumbcacheEntry


class Thumbcache:
    """This class combines the thumbnailindex and thumbcachefile.

    The class looks up all files inside ``path`` that have the same ``prefix``.

    Args:
        path: The directory that contains the thumbcache files.
        prefix: The start of the name to search for.
    """

    def __init__(self, path: Path, prefix: str = "thumbcache") -> None:
        self._mapping: dict[str, Path] = {}
        self.index_file, self.cache_files = self._populate_files(path, prefix)

    def _populate_files(self, path: Path, prefix: str) -> tuple[Path, list[Path]]:
        cache_files = []
        index_file = None
        for file in path.glob(f"{prefix}*"):
            if file.name.endswith("_idx.db"):
                index_file = file
            else:
                cache_files.append(file)
        return index_file, cache_files

    @property
    def mapping(self) -> dict[int, Path]:
        """Looks at the version field in the cache file header."""
        if not self._mapping:
            for file in self.cache_files:
                with file.open("rb") as cache_file:
                    t_file = ThumbcacheFile(cache_file)
                    key = t_file.type
                self._mapping.update({key: file})
        return self._mapping

    def entries(self) -> Iterator[tuple[Path, ThumbcacheEntry]]:
        """Iterates through all the specific entries from the thumbcache files."""
        used_entries = list(self.index_entries())

        for entry in used_entries:
            yield from self._entries_from_offsets(entry.cache_offsets)

    def index_entries(self) -> Iterator[IndexEntry]:
        """Iterates through all the index entries that are in use."""
        with self.index_file.open("rb") as i_file:
            yield from ThumbnailIndex(i_file).entries()

    def _entries_from_offsets(self, offsets: list[int]) -> Iterator[tuple[Path, ThumbcacheEntry]]:
        """Retrieves Thumbcache entries from a ThumbcacheFile using offsets."""
        for idx, offset in enumerate(offsets):
            if offset == 0xFFFFFFFF:
                continue

            cache_path = self.mapping.get(idx)

            with cache_path.open("rb") as cache_file:
                yield cache_path, ThumbcacheFile(cache_file)[offset]



from __future__ import annotations

from argparse import ArgumentParser
from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from dissect.thumbcache.thumbcache import ThumbcacheEntry


def create_argument_parser(name: str) -> ArgumentParser:
    parser = ArgumentParser(name)
    parser.add_argument("cache_path", type=Path)
    parser.add_argument("--output-dir", "-o", type=Path, default=Path())

    return parser


def write_entry(output_path: Path, entry: ThumbcacheEntry, file_prefix: str) -> None:
    output_file = (output_path / f"{entry.hash}/{entry.identifier}_{file_prefix}").with_suffix(".bmp")
    output_file.parent.mkdir(parents=True, exist_ok=True)
    with output_file.open("wb") as output:
        output.write(entry.data)



from __future__ import annotations

from typing import TYPE_CHECKING

from dissect.thumbcache.exceptions import Error
from dissect.thumbcache.thumbcache import Thumbcache
from dissect.thumbcache.tools.utils import create_argument_parser, write_entry

if TYPE_CHECKING:
    from pathlib import Path


def dump_entry_data_through_index(path: Path, output_dir: Path, prefix: str) -> None:
    cache = Thumbcache(path=path, prefix=prefix)
    try:
        for location_path, entry in cache.entries():
            file_name = location_path.stem.split("_", 1)[1]
            write_entry(output_dir, entry, file_name)
    except Error as e:
        print(e)


def main() -> None:
    parser = create_argument_parser("extract indexed entries")
    parser.add_argument(
        "--prefix",
        "-p",
        type=str,
        default="thumbcache",
        help="The file prefix to search for inside the thumbcache directory.",
    )

    args = parser.parse_args()

    path: Path = args.cache_path

    if path.is_dir():
        dump_entry_data_through_index(path, args.output_dir, args.prefix)
    else:
        parser.exit("Please provide the thumbnail cache directory.")


if __name__ == "__main__":
    main()



from __future__ import annotations


class Error(Exception):
    """A generic exception for the thumbcache module."""


class NotAnIndexFileError(Error):
    """Raises if a thumbnail index signature could not be found."""


class InvalidSignatureError(Error):
    """Raises if the signature does not match the expected value."""


class UnknownThumbnailTypeError(Error):
    """Raises if an unknown thumbnail type was found."""



from __future__ import annotations

from dissect.thumbcache.exceptions import Error
from dissect.thumbcache.index import IndexEntry, ThumbnailIndex
from dissect.thumbcache.thumbcache import Thumbcache
from dissect.thumbcache.thumbcache_file import ThumbcacheEntry, ThumbcacheFile

__all__ = [
    "Error",
    "IndexEntry",
    "Thumbcache",
    "ThumbcacheEntry",
    "ThumbcacheFile",
    "ThumbnailIndex",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, BinaryIO

from dissect.thumbcache.c_thumbcache import c_thumbcache
from dissect.thumbcache.exceptions import (
    InvalidSignatureError,
    UnknownThumbnailTypeError,
)
from dissect.thumbcache.util import ThumbnailType, seek_and_return

if TYPE_CHECKING:
    from collections.abc import Iterator

UNKNOWN_BYTES = 8


class ThumbcacheFile:
    __slots__ = [
        "_cached_entries",
        "_entries",
        "_header",
        "fh",
        "offset",
        "signature",
        "size",
        "type",
    ]

    """
    This class defines a thumbcache file, that is usually denoted by
    thumbcache_*.db. Which is is different from the thumbcache_idx.db.

    Args:
        fh: A file-like object.
    """

    _signature = b"CMMM"

    def __init__(self, fh: BinaryIO) -> None:
        self.fh = fh
        self._header = self._get_header_type(self.fh)
        self._cached_entries: dict[int, ThumbcacheEntry] = {}

    def _get_header_type(self, fh: BinaryIO) -> c_thumbcache.CACHE_HEADER_VISTA | c_thumbcache.CACHE_HEADER:
        tmp_header = c_thumbcache.CACHE_HEADER(fh)

        if self._signature != tmp_header.Signature:
            raise InvalidSignatureError(
                f"The signature {tmp_header.Signature!r} does not match the expected {self._signature!r}"
            )

        if tmp_header.Version <= ThumbnailType.WINDOWS_7:
            return c_thumbcache.CACHE_HEADER_VISTA(tmp_header.dumps())
        return tmp_header

    @property
    def header(self) -> c_thumbcache.CACHE_HEADER_VISTA | c_thumbcache.CACHE_HEADER:
        return self._header

    @property
    def version(self) -> ThumbnailType:
        try:
            return ThumbnailType(self.header.Version)
        except ValueError:
            raise UnknownThumbnailTypeError(f"{self.header.Version} is not known.")

    def __getitem__(self, key: int) -> ThumbcacheEntry:
        if key in self._cached_entries:
            return self._cached_entries[key]

        with seek_and_return(self.fh, key) as file:
            item = ThumbcacheEntry(file, self.version)
        self._cached_entries[key] = item
        return item

    def __getattribute__(self, name: str) -> Any:
        try:
            return object.__getattribute__(self, name)
        except AttributeError:
            pass

        return getattr(self.header, name.capitalize())

    def entries(self) -> Iterator[ThumbcacheEntry]:
        with seek_and_return(self.fh, self.fh.tell()):
            try:
                while True:
                    yield ThumbcacheEntry(self.fh, self.version)
            except EOFError:
                pass


class ThumbcacheEntry:
    _signature = b"CMMM"

    """ThumbcacheEntry represents an entry inside the thumbcache file.

    A cache entry contains a hash that is constructed from GUID, NTFS FILEID, the extention and last modified time.
    At least for Windows 7: http://www.swiftforensics.com/2012/06/windows-7-thumbcache-hash-algorithm.html

    Args:
        fh: A reference to binary data, where we can read from.
        type: The type of system that generated the entry.

    Raises:
        InvalidSignatureError: If the entry does not have a proper signature.
    """

    _checksum_lengths = 0x10

    def __init__(self, fh: BinaryIO, type: ThumbnailType) -> None:
        self._type = type
        self._header = self._get_header(type)(fh)

        if self._header.Signature != self._signature:
            raise InvalidSignatureError(
                f"The signature {self._header.Signature!r} does not match the expected {self._signature}."
            )

        additional_bytes = self._checksum_lengths
        if type > ThumbnailType.WINDOWS_7:
            # There are some additional unknown bytes before the header continues,
            # I do not know what kind of information it contains
            fh.read(UNKNOWN_BYTES)
            additional_bytes += UNKNOWN_BYTES

        self.data_checksum: bytes = c_thumbcache.char[8](fh)
        self.header_checksum: bytes = c_thumbcache.char[8](fh)

        self.identifier: str = c_thumbcache.wchar[self._header.IdentifierSize // 2](fh)

        header_size = len(self._header) + self._header.IdentifierSize + additional_bytes

        self._data = fh.read(self._header.Size - header_size)

    def _get_header(
        self, thumbnail_type: ThumbnailType
    ) -> type[c_thumbcache.CACHE_ENTRY_VISTA | c_thumbcache.CACHE_ENTRY]:
        if thumbnail_type == ThumbnailType.WINDOWS_VISTA:
            return c_thumbcache.CACHE_ENTRY_VISTA
        return c_thumbcache.CACHE_ENTRY

    @property
    def hash(self) -> str:
        return self._header.Hash.hex()

    @property
    def extension(self) -> str:
        """This property contains the extension type of the data (Only in VISTA)."""
        if self._type == ThumbnailType.WINDOWS_VISTA:
            return self._header.Extension.strip("\x00")
        return ""

    @property
    def data(self) -> bytes:
        return self._data[self._header.PaddingSize :]

    def __repr__(self):
        return f"{self.hash=} {self.data_checksum=} {self.header_checksum=} {self.identifier=}"



from __future__ import annotations

from dissect.cstruct import cstruct

thumbcache_def = """
struct INDEX_HEADER_V1 {
    char    Signature[4];   // 0x00
    uint32  Version;        // 0x04
    uint32  Unknown1;       // 0x08
    uint32  UsedEntries;    // 0x0C
    uint32  TotalEntries;   // 0x10
    uint32  Unknown2;       // 0x14
}; // 0x18

struct INDEX_HEADER_V2 {
    char    Signature[8];   // 0x00
    uint64  Version;        // 0x08
    uint32  Unknown1;       // 0x10
    uint32  UsedEntries;    // 0x14
    uint32  TotalEntries;   // 0x18
    uint32  Unknown2;       // 0x1B
}; // 0x20

struct VISTA_ENTRY {
    char    Hash[8];        // 0x00
    uint64  LastModified;   // 0x08
    uint32  Flags;          // 0x10
}; // 0x14

struct WINDOWS7_ENTRY {
    char    Hash[8];        // 0x00
    uint32  Flags;          // 0x08
}; // 0x0C

struct WINDOWS8_ENTRY {
    char    Hash[8];        // 0x00
    uint32  Flags;          // 0x08
    uint32  Unknown;        // 0x0C Is sometimes filled with information, couldn't figure out what it meant yet though.
}; // 0x10

struct CACHE_HEADER {
    char    Signature[4];   // 0x00
    uint32  Version;        // 0x04
    uint32  Type;           // 0x08
    uint32  Size;           // 0x0C
    uint32  Offset;         // 0x10
    uint32  Entries;        // 0x14
}; // 0x18

struct CACHE_HEADER_VISTA {
    char    Signature[4];   // 0x00
    uint32  Version;        // 0x04
    uint32  Type;           // 0x08
    uint32  Offset;         // 0x0C
    uint32  Size;           // 0x10
    uint32  Entries;        // 0x14
}; // 0x18

struct CACHE_ENTRY {
    char    Signature[4];   // 0x00
    uint32  Size;           // 0x04
    char    Hash[8];        // 0x08
    uint32  IdentifierSize; // 0x10
    uint32  PaddingSize;    // 0x14
    uint32  DataSize;       // 0x18
    uint32  Unknown;        // 0x1C
}; // 0x20

struct CACHE_ENTRY_VISTA {
    char    Signature[4];   // 0x04
    uint32  Size;           // 0x08
    char    Hash[8];        // 0x0C
    wchar   Extension[4];   // 0x14
    uint32  IdentifierSize; // 0x1C
    uint32  PaddingSize;    // 0x20
    uint32  DataSize;       // 0x24
    uint32  Unknown;        // 0x28
}; // 0x2C
"""
c_thumbcache = cstruct().load(thumbcache_def)



from __future__ import annotations

from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts

from dissect.thumbcache.c_thumbcache import c_thumbcache
from dissect.thumbcache.exceptions import NotAnIndexFileError
from dissect.thumbcache.util import ThumbnailType

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

INDEX_ENTRIES = {
    ThumbnailType.WINDOWS_7: 5,
    ThumbnailType.WINDOWS_81: 11,
    ThumbnailType.WINDOWS_10: 14,
    ThumbnailType.WINDOWS_VISTA: 5,
}

MAX_IMM_OFFSET = 4
BYTES_IN_NUMBER = 4
IDENTIFIER_BYTES = 8


class ThumbnailIndex:
    _signature = b"IMMM"

    def __init__(self, fh: BinaryIO) -> None:
        self.fh = fh
        self._header = None

    @property
    def header(self) -> c_thumbcache.INDEX_HEADER_V1 | c_thumbcache.INDEX_HEADER_V2:
        if self._header is None:
            self._header = self._find_header(self.fh)
        return self._header

    def _find_header(self, fh: BinaryIO) -> c_thumbcache.INDEX_HEADER_V1 | c_thumbcache.INDEX_HEADER_V2:
        """Searches for the header signature, and puts ``fh`` at the correct position.

        From Windows 8.1 onward, the two fields seem to use a 64-bit format field
        inside the header with the value ``0C 00 30 20``.

        Args:
            fh: The file to read the header and indexes from.

        Returns:
            A c_thumbcache.INDEX_HEADER structure.

        Raises:
            NotAThumbnailIndexFileError: If the ``IMMM`` signature could not be found.
        """
        position = fh.tell()
        buffer = fh.read(len(c_thumbcache.INDEX_HEADER_V1))
        offset = buffer.find(self._signature)

        if offset == MAX_IMM_OFFSET:
            fh.seek(position)

            header = c_thumbcache.INDEX_HEADER_V2(fh)
            # From looking at the index files, it has a specific amount of information.
            # It is alligned in the follwing way:
            #   INDEX_HEADER_V2
            #   length of an INDEX_ENTRY of a specific type including the cache_offsets
            # After that point, there are only zero bytes, which seem to have the following relation:
            #   length of the same INDEX_ENTRY - length of the length of INDEX_HEADER_V2
            #
            # TODO: see if it the data contains any useful information.

            # Read one index entry from the file till only zero bytes
            entry = IndexEntry(fh, header.Version)
            _ = entry.header
            _ = entry.cache_offsets

            # Read offset to first entry
            zero_bytes = len(entry.header) + INDEX_ENTRIES.get(header.Version) * BYTES_IN_NUMBER - len(header)
            fh.read(zero_bytes)
            return header
        if offset == 0:
            return c_thumbcache.INDEX_HEADER_V1(buffer)
        raise NotAnIndexFileError(
            f"The index file signature {self._signature!r} could not be found at the expected location."
        )

    @property
    def version(self) -> int:
        return self.header.Version

    @property
    def type(self) -> ThumbnailType:
        return ThumbnailType(self.version)

    @property
    def total_entries(self) -> int:
        return self.header.TotalEntries

    @property
    def used_entries(self) -> int:
        return self.header.UsedEntries

    def entries(self) -> Iterator[IndexEntry]:
        """Returns all index entries that are actually used."""
        for _ in range(self.total_entries):
            entry = IndexEntry(self.fh, self.type)
            _ = entry.header
            _ = entry.cache_offsets

            if entry.in_use():
                yield entry


class IndexEntry:
    def __init__(self, fh: BinaryIO, type: ThumbnailType) -> None:
        self.fh = fh
        self.type = type
        self._header = None
        self._data = None

    @property
    def header(
        self,
    ) -> c_thumbcache.VISTA_ENTRY | c_thumbcache.WINDOWS7_ENTRY | c_thumbcache.WINDOWS8_ENTRY:
        if not self._header:
            self._header = self._select_header()
        return self._header

    def _select_header(
        self,
    ) -> c_thumbcache.VISTA_ENTRY | c_thumbcache.WINDOWS7_ENTRY | c_thumbcache.WINDOWS8_ENTRY:
        """Selects header version according to the thumbnailtype."""
        if self.type == ThumbnailType.WINDOWS_VISTA:
            return c_thumbcache.VISTA_ENTRY(self.fh)
        if self.type == ThumbnailType.WINDOWS_7:
            return c_thumbcache.WINDOWS7_ENTRY(self.fh)
        return c_thumbcache.WINDOWS8_ENTRY(self.fh)

    def in_use(self) -> bool:
        return self.identifier != b"\x00" * IDENTIFIER_BYTES

    @property
    def identifier(self) -> bytes:
        return self.header.Hash

    @property
    def flags(self) -> int:
        return self.header.Flags

    @property
    def cache_offsets(self) -> list[int]:
        """Retrieves the index data entries.

        These are offsets into the thumbcache files, where the order specifies in which of the files.
        More information about the order can be found in :class:`Thumbcache`.

        """
        if not self._data:
            size = INDEX_ENTRIES.get(self.type)
            self._data = c_thumbcache.uint32[size](self.fh)
            if self.type > ThumbnailType.WINDOWS_7:
                # Alignment step
                self.fh.read((size % 2) * BYTES_IN_NUMBER)
        return self._data

    @property
    def last_modified(self) -> datetime:
        if self.type == ThumbnailType.WINDOWS_VISTA:
            return ts.wintimestamp(self._header.LastModified)
        return None

    def __repr__(self) -> str:
        return (
            f"identifier={self.identifier.hex()} flags={hex(self.flags)} "
            f"cache_offsets={[hex(x) for x in self.cache_offsets]}"
        )
